// -*- c++ -*-
// $Id: trapsim.C 223 2007-09-27 01:42:02Z olsonse $
/*
 * Copyright 2005-2007 Spencer Olson
 *
 * $Log$
 *
 */

/** \file
 * A uniform box with periodic (3D) boundary conditions.  Not too exciting.
 */

static char const trapsim_rcsid[] = "$Id: trapsim.C 223 2007-09-27 01:42:02Z olsonse $";



/* for the laserfield.h file. */
#define DISABLE_SHELL_LOOKUP

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <cfloat>
#include <stdexcept>
#include <sys/times.h>
#include <unistd.h>
#include <signal.h> // so we can issue ^C to gently stop the loading

#include <olson-tools/strutil.h>
#include <olson-tools/physical.h>
#include <olson-tools/bfield.h>
#include <olson-tools/random/random.h>
#include <olson-tools/listutil.h>
#include <olson-tools/KeyedBin.h>
#include <olson-tools/GenericBin.h>

#include <dsmc/DSMC.h>
#include <dsmc/ParticleSrc.h>
#include <dsmc/Wall.h>
#if defined(USE_MPI)
#  include <dsmc/ParticleNode.h>
#endif

#include <laser-cooling/StateParticle.h>

#define BIN_VELOCITY 1

static std::vector<std::string> rcsid_vector;
int add_rcsid(std::string rcsid) {
    rcsid_vector.push_back(rcsid);
    return 1;
}

std::string get_rcsid_strings() {
    std::ostringstream ostr;
    ostr << "# ";
    std::copy(rcsid_vector.begin(),
              rcsid_vector.end(),
              std::ostream_iterator<std::string>(ostr, "\n# "));
    ostr << "end of rcsid list";
    return ostr.str();
}

int trapsim_dummy = add_rcsid(trapsim_rcsid);

static bool quit_requested = false;
void quitRequested(int inSignal) {
    quit_requested = true;
}

using namespace physical::constants;
using namespace physical::units;

const double temperature = 10.*uK;

#ifdef BIN_VELOCITY
GenericBin<double,1000> vbin[3] = {
    GenericBin<double,1000>(-100*cm/s,100*cm/s),
    GenericBin<double,1000>(-100*cm/s,100*cm/s),
    GenericBin<double,1000>(-100*cm/s,100*cm/s)
};

GenericBin<double,1000> tmpvbin[3] = {
    GenericBin<double,1000>(-100*cm/s,100*cm/s),
    GenericBin<double,1000>(-100*cm/s,100*cm/s),
    GenericBin<double,1000>(-100*cm/s,100*cm/s)
};
#endif


#  define P_XYZ   Gaussian3DParticleDistrib
#  define PV_XYZ  Gaussian3Distros

#include "../common/selectlaser.h"


int camera_F2 = 0;
int camera_F3 = 0;
GenericBin<double,1000> camera_bin(camera_min[X],camera_max[X]);


typedef DSMC< RK5AdaptiveIntegrator< StatisticalForceRKWrapper<LaserForce> >,
          octree::CMAndVolNotShrink,
          NULLDXHook,
          Particle_t > DSMC_ALG;

int mainprog(int argc, char *argv[]) {
    /* need to be able to add particles to the system. */
    int8_t particle_type = getParticleTypeIndex("85Rb");

    LaserForce laserForce;
    laserForce.super0::mass = particle_info[particle_type].mass;
    initLaser(laserForce);


    const double N              = 20e6; // initial number of MOT atoms
    const double sigma_x        = initial_particle_sigma_x(laserForce); // MOT size

    const double n0             = N / CUBE(sqrt(M_PI * 2) * sigma_x );
    const double lambda_mfp     = 1.0 / (M_SQRT2 * particle_info[particle_type].cross_section * n0);
    const double Kn             = lambda_mfp / sigma_x;
    /* We'll ensure that dt ~= min(0.1 Coll.Time,0.1 1m/v_avg)  */
    static double dt_max =
        0.1 /* we need to allow excursions 1/10 the size of L */
        * sigma_x /* characteristic length. */
        / sqrt( 8.0*K_B * temperature / pi / particle_info[particle_type].mass);
    const double dt_Kn =
        0.1 /* we need to allow excursions 1/10 the size of L */
        * lambda_mfp /* mean free path. */
        / sqrt( 8.0*K_B * temperature / pi / particle_info[particle_type].mass);
    const double dt = 0.1*std::min(dt_Kn, dt_max);
    static double t_max = 200.*ms + dt /* dt * 200.0 */;
    static int spit_out_data = 0;

    const double desired_samples = 115;
    const double sample_wait_time = 100.0*s; /* time before real sampling happens. */
    static int sample_interval = int( (t_max/*-sample_wait_time*/) / ( desired_samples * dt ) );
    static int write_interval = int(1*ms/dt) ;//* sample_interval;
    int write_particles_interval = write_interval; /* must be multiple of write_interval */


    ParticleSrc< P_XYZ, PV_XYZ, Particle_t > atomsrc;
    atomsrc.seed.type = particle_type;
    atomsrc.seed.dt   = 10*us;
    atomsrc.seed.state = StateParticle::F2;

#ifndef P_XYZ_FLAT
    atomsrc.pos_distrib.sigma_x         = sigma_x;
#else
    atomsrc.pos_distrib.delta_x         = sigma_x;
#endif

    atomsrc.flux                        = N / dt;

#if defined(USE_MPI)
    particle_info[particle_type].size   = 10;
#else
    particle_info[particle_type].size   = atomsrc.flux * 1e-5 * dt;
#endif
    atomsrc.pulsed_source               = true;
    atomsrc.pulse_duration              = dt;
    atomsrc.origin                      = V3(0,0,0);
    atomsrc.temperature                 = temperature;
    atomsrc.velocity                    = 0.0; //200.0*m/s;
    //atomsrc.velocity[X]                 = 20.0*cm/s;
    atomsrc.initialize();


    std::cout << get_rcsid_strings()                            << '\n'
              << "(" << getpid() << ")  t_max address:  " << (&t_max) << '\n'
              << "(" << getpid() << ")  spit_out_data address:  " << (&spit_out_data) << '\n'
              << "(" << getpid() << ")  sample_interval address:  " << (&sample_interval) << '\n'
              << "(" << getpid() << ")  write_interval address:  " << (&write_interval) << '\n'
              << "(" << getpid() << ")  write interval (s) :  " << (write_interval*dt) << '\n'
              << "(" << getpid() << ")  dt:  " << dt << " s\n"
              << "(" << getpid() << ")  N :  " << N << '\n'
              << "(" << getpid() << ")  n0:  " << (n0/CUBE(cm)) << " /cm^3 \n"
              << "(" << getpid() << ")  Kn:  " << Kn << '\n'
              << "(" << getpid() << ")  lambda_mfp:  " << (lambda_mfp/mm) << " mm\n"
              << "(" << getpid() << ")  FN:  " << particle_info[particle_type].size << std::endl;
    printLaserInfo(laserForce, std::cout, "(" + to_string(getpid()) + ")  ");

#if defined(USE_MPI)
    ParticleNode<Particle_t,DSMC_ALG> node;
    DSMC_ALG & dsmc = node.container;

    /* we want the MPI divisions done in the X-direction. */
    node.physical_orientation = X;
    node.amin_N_per_proc = 1500.0;

    /* here goes nothing... */
    node.init(argc, argv);
#else
    DSMC_ALG dsmc;
#   define node dsmc
#endif

    setLogProgramName("trapsim");
    std::string logfile("logger.out");
    if (logfile.length() > 0) {
        setLogFile(logfile.c_str());
    }

    dsmc.integrator.derivs = (derivativesFunction)Derivs<LaserForce>::derivs;
    dsmc.integrator.errmax = 1e-6;
    dsmc.integrator.rkTweak.statisticalForcePtr = &laserForce;
    dsmc.allow_collisions = false;

    addBoundaries(dsmc);

    dsmc.data_radius_of_influence = 4;
    dsmc.minimum_octree_particles = 20;

    /* ** plotting all particles. */
    std::ofstream fout("particles.plt");
    std::ofstream fdata("data.plt");
    std::ofstream fpout("data.dat");

    std::ofstream fcout("collision-totals.dat");
    fout << "set xlabel 'x'\nset ylabel 'y'\nset zlabel 'z'\n";
    std::string tree_plotopt = "w lp ps 1 t ''" ;
    std::string part_plotopt = "w p pt 6 ps 2 lt 2 t 'particles'";
    std::ostringstream ostr;


    fcout << "#  Collisions Totals\n"
             "#  Sim. Time(s)        inst. coll./s"
             "          avg. coll./s"
             "          avg. coll. selection rate (per atom)"
             "          Num. Super Particles"
             "          Cycle CPU Time(s)\n";


    std::ofstream vbinout[3] = {
        std::ofstream("vbinX.dat"),
        std::ofstream("vbinY.dat"),
        std::ofstream("vbinZ.dat")
    };

    std::ofstream scatout("scatter.dat");
    scatout.precision(10);

    std::ofstream cameraout("camera.dat");

    std::ofstream state_lifetime_camera("state_lifetime_camera.dat");

    double max_dx[3] = {0.0, 0.0, 0.0};

    /* now let the dsmc have some fun with the particles. */
    try {
#if defined(USE_MPI)
        /* another hack because I ran out of memory somewhere. spread the load */
        if (atomsrc.pulsed_source && atomsrc.pulse_duration < (1.5*dt)) {
#if 0
            const double ddt = 0.1 * dt;
            for (double tt = ddt; tt < (dt*1.0001); tt += ddt)
#else
#  define ddt dt
#endif
            {
                if (node.mpi_info.arank >= 0) {
                    node.addParticles(atomsrc.createParticles(ddt));
                }

                /* only the rank=0 node is active now.  Thus, we decompose right
                 * away. */
                node.decomposeEntireDomain();
            }
        }
#endif

        signal(SIGINT, quitRequested);
        double seconds_per_clock_tick = 1.0 / sysconf(_SC_CLK_TCK);
        while ( !quit_requested && dsmc.getTotalTime() <= t_max) {

#if defined(USE_MPI)
            if (node.mpi_info.arank >= 0)
#endif
            {
                node.addParticles(atomsrc.createParticles(dt));
            }

            set_laser_timing(laserForce, dsmc.getTotalTime());

            struct tms ti, tf;
            times(&ti);
            node.evolve(dt, &laserForce);
            times(&tf);

            if (dsmc.getTotalTime() >= sample_wait_time) spit_out_data = 1;

            {
                double cycle_cpu_time = 
                     ( ( (tf.tms_utime + tf.tms_stime) - (ti.tms_utime + ti.tms_stime) )
                       * seconds_per_clock_tick
                     );

                /* plot collision totals every time-step. */
                fcout << dsmc.getTotalTime()       << '\t'
                      << (dsmc.getCycleCollisions()/(dt*dsmc.numberParticles())) << '\t'
                      << (dsmc.getTotalCollisions()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                      /* the following is supposed to be collision selection rate
                       * per atom.  That is why the factor of 2 precedes the line.
                       */
                      << (2*dsmc.getTotalCollisionSelections()/(dsmc.getTotalTime()*dsmc.numberParticles())) << '\t'
                      << dsmc.numberParticles() << '\t'
#if defined(USE_MPI)
                      << node.cpu_time << '\t'
#endif
                      << cycle_cpu_time << std::endl;
            }

            if ((int(dsmc.getTotalTime() / dt) - 1) % sample_interval  == 0) {

#ifdef BIN_VELOCITY

                for (int i = 0; i < dsmc.numberParticles(); i++) {
                    Vector<double,3> & v = V3C(dsmc.getParticles()[i]->vel);
                    vbin[X].bin(v[X]);
                    vbin[Y].bin(v[Y]);
                    vbin[Z].bin(v[Z]);
                }
#endif


            }

            /* **** TAKE CARE OF PLOTTING ALL PARTICLES. */
            if (dsmc.getTotalTime() >= (t_max-dt) || (int(dsmc.getTotalTime() / dt) - 1) % write_interval  == 0) {
                std::string tot_time = to_string(dsmc.getTotalTime());
#if 1
                /* note that this octree is not necessarily representative of the
                 * current particles.  This is because there has been a call to
                 * dsmc.evolve since the Octree was built.
                 */
#if 0
                dsmc.plotParticlesAndTree(fout,
                                          tot_time + " s",
                                          tree_plotopt,
                                          part_plotopt)
    	            << "pause p\n" << std::flush;
#endif
                dsmc.printParticles(fout) << "\n\n" << std::flush;
#endif

                camera_F2 = camera_F3 = 0;
                camera_bin.clearBins();
                for (int i = 0; i < dsmc.numberParticles(); i++) {
                    StateParticle & p = *dsmc.getParticles()[i];
                    if (V3C(p.pos) >= camera_min && V3C(p.pos) <= camera_max) {
                        if(p.state == StateParticle::F2) camera_F2++;
                        else camera_F3++;

                        /* The camera integrates in the Y direction. */
                        camera_bin.bin(p.pos[X]);
                    }
                }

                state_lifetime_camera
                    << tot_time << '\t'
                    << camera_F2 << '\t'
                    << camera_F3 << std::endl;

                camera_bin.print(cameraout, tot_time + "\t") << std::endl;

#ifdef SPONTSCATMON_H
                LaserCooling::SScatBin *= (1.0/(write_interval * dt));
                LaserCooling::SScatBin.print(scatout);
                LaserCooling::SScatBin.clearBins();
#endif


                if (dsmc.getInterpolator()) {
                    /* ok, let's spit out the data tree. */
#if 0
                    dsmc.getInterpolator()->
                         plotParticlesAndTree(fdata,
                                              "DSMC Data Tree:  " + tot_time + " s",
                                              tree_plotopt,
                                              part_plotopt)
                     << "pause -1\n" << std::flush;
#endif

#if 0
                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    dsmc.getInterpolator()->printParticles(fpout) << "\n\n";
#endif

#ifdef BIN_VELOCITY
                    /* We're going to normalize the bin before we write it
                     * out. */
                    int vbinsum[3] = {0,0,0};
                    for (unsigned int i = 0; i < vbin[X].nBins(); i++) {
                        vbinsum[X] += vbin[X].bins[i];
                        vbinsum[Y] += vbin[Y].bins[i];
                        vbinsum[Z] += vbin[Z].bins[i];
                    }

                    /* the factor of 1e7 is just to make sure that we have a
                     * dynamic range of 0:1e4 after converting back to
                     * integer. (Remember that there are 1000 bins in the
                     * histogram.) */
                    for (unsigned int i = 0; i < vbin[X].nBins(); i++) {
                        tmpvbin[X].bins[i] = (int)rint( 1e7 * ((double)vbin[X].bins[i]) / ((double)vbinsum[X]) );
                        tmpvbin[Y].bins[i] = (int)rint( 1e7 * ((double)vbin[Y].bins[i]) / ((double)vbinsum[Y]) );
                        tmpvbin[Z].bins[i] = (int)rint( 1e7 * ((double)vbin[Z].bins[i]) / ((double)vbinsum[Z]) );
                    }

                    /* write out new shmooshed together spline data. */
                    vbinout[X] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    vbinout[Y] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    vbinout[Z] << "#  Normalized velocity histogram\n#  Time:  " << tot_time << " s\n";
                    tmpvbin[X].print(vbinout[X],tot_time + '\t') << "\n\n" << std::flush;
                    tmpvbin[Y].print(vbinout[Y],tot_time + '\t') << "\n\n" << std::flush;
                    tmpvbin[Z].print(vbinout[Z],tot_time + '\t') << "\n\n" << std::flush;
#endif

                    if ( !spit_out_data ) {
                        /* we'll just keep clearing the bins until we need to
                         * really start time-averaging.
                         */
#ifdef BIN_VELOCITY
                        vbin[X].clearBins();
                        vbin[Y].clearBins();
                        vbin[Z].clearBins();
#endif
                    }

                    /* now let's zero the collisions-per-octree-node counters
                     * so that we only get number collisions per sample time.
                     */
                    DSMCData<Particle_t> ** d = dsmc.getInterpolator()->getParticles();
                    for (int i = dsmc.getInterpolator()->numberParticles() - 1; i >= 0; i--) {
                        d[i]->data.name.collisions = 0;
                    }
                } else if (dsmc.numberParticles() > 0 && spit_out_data) {
                    /* we have to make our own interpolator since DSMC didn't need
                     * to construct it. */
                    octree::OctreeInterpolator<DSMCData<Particle_t>,Particle_t> ointerp;
                    DSMCData<Particle_t>::InitInfo init = {dsmc.getParticles()};

                    ointerp.createDataOctree(dsmc.getOctree(), init);
                    ointerp.setRadiiOfInfluence(dsmc.data_radius_of_influence);

                    fpout << "#  Data Particles\n#  Time:  " << tot_time << " s\n";
                    ointerp.printParticles(fpout) << "\n\n";
                }

            }/* if */

            for (int i = X; i <= Z; i++)
                max_dx[i] = std::max(max_dx[i], dsmc.max_particle_transit(i));
        }/* while */
    } catch (std::exception & e) {
        std::cout << "exception:  " << e.what() << std::endl;
        char cmd[1024] = "";
        sprintf(cmd, "pmap %d", getpid());
        system(cmd);
        fflush(stdout);
        fflush(stderr);
    }

    fout.close();
    fpout.close();
    fdata.close();
    fcout.close();

    vbinout[X].close();
    vbinout[Y].close();
    vbinout[Z].close();

    std::cout << "max Particle Transit:  ("
              << max_dx[X] << ','
              << max_dx[Y] << ','
              << max_dx[Z] << ')'
              << std::endl;

    return 0;
}

int main(int argc, char *argv[]) {
    try {
        return mainprog(argc,argv);
    } catch (std::exception &e) {
        log_severe("%d:  exception caught:  %s", getpid(), to_string(e.what()).c_str());
    } catch (...) {
        log_severe("%d:  non-std:: exception caught.", getpid());
    }

    return EXIT_FAILURE;
}
