// -*- c++ -*-
// $Id$
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log$
 */


#ifndef CONSTANTLASERFIELD_H
#define CONSTANTLASERFIELD_H

#define P_XYZ_FLAT 1

#include <olson-tools/Vector.h>
#include <dsmc/BoundaryObject.h>
#include <laser-cooling/OpticalForce.h>
#include <laser-cooling/SpontScatMon.h>
//typedef Particle Particle_t;
typedef StateParticle Particle_t;

using LaserCooling::OpticalForce;

class ConstantLaserField : public virtual BaseField {
  public:
    typedef BaseField super0;

    /** Default constructor. */
    ConstantLaserField(const double & _I_0 = physical::units::mW / SQR(physical::units::cm),
                       const Vector<double,3> _k  = V3(1,0,0),
                       const double & _lambda     = 780.244 * physical::units::nm) :
        super0(),
        I_0(_I_0),
        k(_k/_k.abs()),
        lambda(_lambda){ }

    double I_0;                 /* Intensity at focus (W/m^2). */
    Vector<double,3> k;         /* wave number (normalized) */
    double lambda;              /* wavelength of laser (m) */

    /** calculates and returns the intensity of the gaussian beam. */
    inline double operator()(const Vector<double,3> & r) const {
        return I_0;
    }

    /** Calculates the magnitude of the field.
     *   E(rho) = E_0 * exp( -rho^2 / w^2) * sqrt( w0 / w )
     */
    inline double E(const Vector<double,3> & r) const {
        return sqrt(operator()(r));
    }
};

typedef OpticalForce< ConstantLaserField, 0, SpontScatMonitor > beam;

typedef AddForce< beam, Gravity > LaserForce;

void initLaser(LaserForce & laserForce) {
    laserForce.beam::laser.k  = V3(1,0,0);
    laserForce.beam::laser.lambda = 779.244 * nm;
    laserForce.beam::laser.I_0 = 500*mW/SQR(cm);
    laserForce.beam::detuning = 0;
    laserForce.beam::initialize();
    laserForce.Gravity::bg = 0.0;
    initSScatBin();
}

double totalIntensity(const LaserForce & laserForce, const Vector<double,3> & r) {
    return laserForce.beam::laser(r);
}

std::ostream & printLaserInfo(const LaserForce & laserForce, std::ostream & out, const std::string & prefix = "") {
    double s0 = (laserForce.laser.I_0/(1.6*mW/SQR(cm)));
    out << prefix << "detuning:    " << laserForce.detuning << std::endl;
    out << prefix << "I_0:       " << laserForce.laser.I_0 << std::endl;
    out << prefix << "gamma_p(v=0):  " << 
    ( (pi * 6*MHz) * s0 / ( 1.0 + s0 + SQR(2.0*laserForce.detuning/(6*MHz)) ) )
        << std::endl;

    return out;
}

double initial_particle_sigma_x(const LaserForce & laserForce) {
    return 10.*mm;
}


static const struct {
/* LASER TEST RANGES */
static const double
    xi = -10.0*mm,
    xf =  10.0000001*mm,
    dx =  0.5*mm,

    yi = -10.0*mm,
    yf =  10.0000001*mm,
    dy =  0.5*mm,

    zi = -10.0*mm,
    zf =  10.00000001*mm,
    dz =   .5*mm;

static const double Imax = 2e6; /* W/m^2 */
} tr;



/* ************ BEGIN DSMC BOUNDARIES ***************** */

const double Len = 20.0*mm;

typedef octree::TeleporterBoundaryObject<Particle_t> tele_type;
const int num_tele = 6;
tele_type tele[num_tele] = {
    /*        normal     point             side-length  scaling     teleport-distance */
    tele_type(V3(-1,0,0), V3( 0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm, V3(-0.9999*Len,0,0)),
    tele_type(V3( 1,0,0), V3(-0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm, V3( 0.9999*Len,0,0)),
    tele_type(V3(0,-1,0), V3(0, 0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm, V3(0,-0.9999*Len,0)),
    tele_type(V3(0, 1,0), V3(0,-0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm, V3(0, 0.9999*Len,0)),
    tele_type(V3(0,0,-1), V3(0,0, 0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm, V3(0,0,-0.9999*Len)),
    tele_type(V3(0,0, 1), V3(0,0,-0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm, V3(0,0, 0.9999*Len))
};

template <class DSMC>
void addBoundaries(DSMC & dsmc) {
    for (int i = 0; i < num_tele; i++) {
        dsmc.addBoundaryObject(&tele[i]);
    }
}

/* ************** END DSMC BOUNDARIES ***************** */

#endif // CONSTANTLASERFIELD_H
