// -*- c++ -*-
// $Id$
/*
 * Copyright 2004 Spencer Olson
 *
 * $Log$
 */


#ifndef SPHERICALHARMONICLASER_H
#define SPHERICALHARMONICLASER_H

#include "OpticalForce.h"
#include "SpontScatMon.h"

using LaserCooling::OpticalForce;

class SphericalHarmonicLaser : public virtual BaseField {
  public:
    typedef BaseField super0;

    /** Default constructor. */
    SphericalHarmonicLaser(const double & _I_0 = physical::units::mW / SQR(physical::units::cm),
                  const double & _w0  = 300.0*physical::units::um,
                  const Vector<double,3> _k  = V3(1,0,0),
                  const Vector<double,3> _r0 = V3(0,0,0),
                  const double & _lambda     = 780 * physical::units::nm) :
        super0(),
        I_0(_I_0),
        w0(_w0),
        k(_k/_k.abs()),
        lambda(_lambda),
        r0(_r0) { }

    double I_0;                 /* Intensity at focus (W/m^2). */
    double w0;                  /* I(w0) = I_0. */
    Vector<double,3> k;         /* wave number (normalized) */
    double lambda;              /* wavelength of laser (m) */
    Vector<double,3> r0;        /* location of focal spot (m) */

    /** calculates and returns the intensity of the gaussian beam. */
    inline double operator()(const Vector<double,3> & r) const {
        return I_0 * SQR(r - r0) / SQR(w0);
    }

    /** Calculates the magnitude of the field.
     *   E(rho) = E_0 * exp( -rho^2 / w^2) * sqrt( w0 / w )
     */
    inline double E(const Vector<double,3> & r) const {
        return sqrt(operator()(r));
    }
};

typedef OpticalForce< SphericalHarmonicLaser, 0, SpontScatMonitor > beam;

typedef AddForce< beam, Gravity > LaserForce;

void initLaser(LaserForce & laserForce) {
    laserForce.beam::laser.k  = V3(1,0,0);
    laserForce.beam::laser.lambda = 779.244 * nm;
    laserForce.beam::laser.I_0 = 500*mW/SQR(cm);
    laserForce.beam::laser.w0 = 300*um;
    laserForce.beam::detuning = 0;
    laserForce.beam::initialize();
    laserForce.Gravity::bg = 0.0;
    initSScatBin();
}

double totalIntensity(const LaserForce & laserForce, const Vector<double,3> & r) {
    return laserForce.beam::laser(r);
}

std::ostream & printLaserInfo(const LaserForce & laserForce, std::ostream & out, const std::string & prefix = "") {
    out << prefix << "detuning:    " << laserForce.detuning << std::endl;
    out << prefix << "w0:       " << laserForce.laser.w0 << std::endl;
    out << prefix << "I_0:       " << laserForce.laser.I_0 << std::endl;
    out << prefix << "Trap Frequency:  " << 
    (sqrt(  (h * 6*MHz / (4. * 85.*amu) )
         * (6*MHz/laserForce.detuning)
         * (laserForce.laser.I_0/(1.6*mW/SQR(cm)))
         / SQR(laserForce.laser.w0)
        ) *0.5 / pi) << std::endl;

    return out;
}


static const struct {
/* LASER TEST RANGES */
static const double
    xi = -105*um,
    xf =  105.0000001*um,
    dx =  0.25*um,

    yi = -105*um,
    yf =  105.0000001*um,
    dy =  0.25*um,

    zi = -10.75*mm,
    zf =  10.7500000001*mm,
    dz =   .5*mm;

static const double Imax = 2e6; /* W/m^2 */
} tr;

#endif // SPHERICALHARMONICLASER_H
