
#ifndef SINGLE_LASERFIELD_H
#define SINGLE_LASERFIELD_H

#include <olson-tools/Vector.h>
#include <dsmc/BoundaryObject.h>
#include <laser-cooling/OpticalForce.h>
#include <laser-cooling/SpontScatMon.h>
#include <laser-cooling/ScaleIntensity.h>

//typedef Particle Particle_t;
typedef StateParticle Particle_t;


using LaserCooling::OpticalForce;
using LaserCooling::LaserFieldLookup;
using LaserCooling::IntensityRecord;
using LaserCooling::SpontScatMonitor;
using LaserCooling::ScaleIntensity;

typedef OpticalForce<
    ScaleIntensity< LaserFieldLookup< 2, AxiSymFieldLookup< IntensityRecord<2> > > >,
    0,
    SpontScatMonitor
 > beam;

typedef AddForce< beam, Gravity > LaserForce;

void initLaser(LaserForce & laserForce) {
    /* set the laser timing. */
    laserForce.beam::laser.t_i = 0.0;
    laserForce.beam::laser.t_f = 5.*ms;
    laserForce.beam::laser.set_scale(0.0);

    laserForce.beam::laser.k  = V3(1,0,0);
    laserForce.beam::laser.lambda = 775.244 * nm;
    laserForce.beam::laser.setPowerFactor(200*mW, 0.89 * SQR(cm));
    laserForce.beam::detuning = 0;
    laserForce.beam::laser.readindata("/tmp/testlaser/lg0001_Frac0790_Gauss1710_I0000_F0200.dat");
    laserForce.beam::laser.rotateField(laserForce.beam::laser.k);
    laserForce.beam::initialize();
    LaserCooling::initSScatBin();
}

double totalIntensity(const LaserForce & laserForce, const Vector<double,3> & r) {
    return laserForce.beam::laser(r);
}

std::ostream & printLaserInfo(const LaserForce & laserForce, std::ostream & out, const std::string & prefix = "") {
    out << prefix << "detuning:    " << laserForce.detuning << std::endl;
    out << prefix << "Power:       " << laserForce.laser.P << std::endl;
    out << prefix << "PowerFactor: " << laserForce.laser.PowerFactor << std::endl;

    return out;
}

double initial_particle_sigma_x(const LaserForce & laserForce) {
    return 300*um;
}

void set_laser_timing(LaserForce & laserForce, const double & t) {
    laserForce.laser.set_scale(t);
}

static const struct {
/* LASER TEST RANGES */
static const double
    xi = -50*um,
    xf =  50*um,
    dx =  0.5*um,

    yi = -50*um,
    yf =  50*um,
    dy =  0.5*um,

    zi = -10*mm,
    zf =  10*mm,
    dz =   .2*mm;

static const double Imax = 2e6; /* W/m^2 */
} tr;



/* ************ BEGIN DSMC BOUNDARIES ***************** */

const double Len = 20.0*mm;

typedef octree::VacuumBoundaryObject<Particle_t>     vac_type;
/** Vacuum boundary conditions. */
const int num_vacuo = 6;
vac_type vacuo[num_vacuo] = {
    /*        normal     point            side-length    scaling */
    vac_type(V3(-1,0,0), V3( 0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm),
    vac_type(V3( 1,0,0), V3(-0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm),
#if 1
    vac_type(V3(0,-1,0), V3(0, 0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0, 1,0), V3(0,-0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0,0,-1), V3(0,0, 0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm),
    vac_type(V3(0,0, 1), V3(0,0,-100.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm)
#else
    vac_type(V3(0,-1,0), V3(0, 70.*um,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0, 1,0), V3(0,-70.*um,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0,0,-1), V3(0,0, 70.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm),
    vac_type(V3(0,0, 1), V3(0,0,-70.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm)
#endif
};

template <class DSMC>
void addBoundaries(DSMC & dsmc) {
    for (int i = 0; i < num_vacuo; i++) {
        dsmc.addBoundaryObject(&vacuo[i]);
    }
}

/* ************** END DSMC BOUNDARIES ***************** */
#endif // SINGLE_LASERFIELD_H
