
#ifndef SINGLE_GAUSSIAN_H
#define SINGLE_GAUSSIAN_H

#include <olson-tools/Vector.h>
#include <dsmc/BoundaryObject.h>
#include <laser-cooling/OpticalForce.h>
#include <laser-cooling/SpontScatMon.h>
//typedef Particle Particle_t;
typedef StateParticle Particle_t;


using LaserCooling::GaussianLaser;

typedef OpticalForce< GaussianLaser > beam;

typedef AddForce< beam, Gravity > LaserForce;

void initLaser(LaserForce & laserForce) {
    laserForce.beam::laser.k  = V3(0,0,1);
    laserForce.beam::laser.lambda = 780.244 * nm;
    laserForce.beam::detuning = -10*MHz;
    laserForce.beam::laser.setPower(2*mW);
    laserForce.beam::laser.setFocalProperties(pi*1*mm, 20*cm);
    laserForce.beam::initialize();
}

double totalIntensity(const LaserForce & laserForce, const Vector<double,3> & r) {
    return laserForce.beam::laser(r);
}

std::ostream & printLaserInfo(const LaserForce & laserForce, std::ostream & out, const std::string & prefix = "") {
    out << prefix << "w0:  " << laserForce.laser.w0 << std::endl;
    out << prefix << "zR:  " << laserForce.laser.zR << std::endl;
    out << prefix << "detuning:  " << laserForce.detuning << std::endl;

    return out;
}

double initial_particle_sigma_x(const LaserForce & laserForce) {
    return laserForce.laser.w0;
}

static const struct {
/* LASER TEST RANGES */
static const double
    xi = -2*mm,
    xf =  2*mm,
    dx = .05*mm,

    yi = -2*mm,
    yf =  2*mm,
    dy = .05*mm,

    zi = -20*cm,
    zf =  20*cm,
    dz =   5*mm;

static const double Imax = 2e6; /* W/m^2 */
} tr;



/* ************ BEGIN DSMC BOUNDARIES ***************** */

const double Len = 20.0*mm;

typedef octree::VacuumBoundaryObject<Particle_t>     vac_type;
/** Vacuum boundary conditions. */
const int num_vacuo = 6;
vac_type vacuo[num_vacuo] = {
    /*        normal     point            side-length    scaling */
    vac_type(V3(-1,0,0), V3( 0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm),
    vac_type(V3( 1,0,0), V3(-0.5*Len,0,0), V3(0,DBL_MAX,DBL_MAX), 0.5*mm),
#if 1
    vac_type(V3(0,-1,0), V3(0, 0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0, 1,0), V3(0,-0.5*Len,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0,0,-1), V3(0,0, 0.5*Len), V3(DBL_MAX,DBL_MAX,0), 0.5*mm),
    vac_type(V3(0,0, 1), V3(0,0,-100.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm)
#else
    vac_type(V3(0,-1,0), V3(0, 70.*um,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0, 1,0), V3(0,-70.*um,0), V3(DBL_MAX,0,DBL_MAX), 0.5*mm),
    vac_type(V3(0,0,-1), V3(0,0, 70.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm),
    vac_type(V3(0,0, 1), V3(0,0,-70.*um), V3(DBL_MAX,DBL_MAX,0), 0.5*mm)
#endif
};

template <class DSMC>
void addBoundaries(DSMC & dsmc) {
    for (int i = 0; i < num_vacuo; i++) {
        dsmc.addBoundaryObject(&vacuo[i]);
    }
}

/* ************** END DSMC BOUNDARIES ***************** */
#endif // SINGLE_GAUSSIAN_H
