// -*- c++ -*-
// $Id$
/** \file
 * Information for an atomic resonance.
 *
 * Copyright 2004-2007 Spencer Olson
 *
 */

#ifndef ATOMICRESONANCE_H
#define ATOMICRESONANCE_H

#include <olson-tools/physical.h>
#include <olson-tools/power.h>

namespace LaserCooling {

class AtomicResonance {
  public:
    /** AtomicResonance constructor.
     * This defaults to \f$^{87}{\rm Rb}\f$ laser cooling transitions (780.244nm, 6MHz, ...).
     */
    AtomicResonance(double l = 780.244 * physical::unit::nm,
             double G = 6 * physical::unit::MHz,
             double Is = 1.6 * physical::unit::mW / SQR(physical::unit::cm)) {
        set(l,G,Is);
    }

    void set(double l = 780.244 * physical::unit::nm,
             double G = 6 * physical::unit::MHz,
             double Is = 1.6 * physical::unit::mW / SQR(physical::unit::cm))
    {
        _lambda = l;
        _Gamma  = G;
        _Isat   = Is;

        double k = 2.0 * physical::constant::pi / _lambda;

        _s1 = physical::constant::mu_B / (physical::constant::h * _Gamma);
        _s2 = 1.0 / (_lambda * _Gamma);
        _s3 = 0.5 * (physical::constant::h * _Gamma) * k;
    }

  private:

    /** Wavelength in meters. */
    double _lambda;

    /** Linewidth in Hz. */
    double _Gamma;

    /** Saturation intensity in Watts / meter^2. */
    double _Isat;


    /** for computing MOT forces. */
    double _s1; /* s1 = \mu_B / ( h * \Gamma ) */
    double _s2; /* s2 = 1 / ( \lambda * \Gamma ) */
    double _s3; /* s3 = (1/2) * ( h * \Gamma ) * ( 2*\pi )/lambda */

  public:
    const double & lambda() const {return _lambda;}
    const double & Gamma() const {return _Gamma;}
    const double & Isat() const {return _Isat;}
    const double & s1() const {return _s1;}
    const double & s2() const {return _s2;}
    const double & s3() const {return _s3;}
};

} /* namespace LaserCooling */

#endif // ATOMICRESONANCE_H
