// -*- c++ -*-
// $Id$
/** \file
 * Laguerre-Gaussian laser definition.
 *
 * Copyright 2004-2007 Spencer Olson
 *
 */


#ifndef LAGUERREGAUSSIANLASER_H
#define LAGUERREGAUSSIANLASER_H

namespace LaserCooling {

/** Laguere-Gaussian laser class.
 * FIXME:  The focusing on this class may not be correct.
 */
class LaguerreGaussianLaser : public virtual BaseField {
  public:
    typedef BaseField super0;

    /** Default constructor. */
    LaguerreGaussianLaser(const double & _P_0 = physical::units::mW,
                          const double & _w0  = physical::units::cm,
                          const double & _polarization  = 1,
                          const double & _cutoff_radius = DBL_MAX,
                          const Vector<double,3> _k  = V3(1,0,0),
                          const Vector<double,3> _r0 = V3(0,0,0),
                          const double & _lambda     = 780 * physical::units::nm,
                          const double & _zR         = 0.0,
                          const int    & _l = 1,
                          const int    & _p = 0 ) :
        super0(),
        P_0(_P_0),
        w0(_w0),
        polarization(_polarization),
        cutoff_radius2(DBL_MAX),
        k(_k/_k.abs()),
        r0(_r0),
        lambda(_lambda),
        zR(_zR),
        l(_l),
        p(_p) {
        if (_cutoff_radius < DBL_MAX) {
            cutoff_radius2 = SQR(_cutoff_radius);
        }
        if (zR == 0.0) zR = zR_collimated();
    }

    double P_0;                 /* Power of beam (W). */
    double w0;                  /* beam waist (m). */
    double polarization;        /* Polarization (+1 or -1; see above explanation. */
    double cutoff_radius2;      /* cutoff radius (m) (due to windows etc.) */
    Vector<double,3> k;         /* wave number (normalized) */
    Vector<double,3> r0;        /* location of focal spot (m) */
    double lambda;              /* wavelength of laser (m) */
    double zR;                  /* Rayleigh Range (m). Defaults to collimated. */
    int p;
    int l;

    /** Calculates and sets the Rayleigh range (zR) from f-Number.
     * \f$ z_R = 2 * w_0 (f/#) \f$.
     */
    void set_zR_fromFNum(const double & FNUM) {
        zR = 2.0 * w0 * FNUM;
    }

    /** Calculates and sets the Rayleigh range (zR) from numerical aperture.
     * \f$ z_R = w_0 / NA \f$.
     */
    void set_zR_fromNA(const double & NA) {
        zR = w0 / NA;
    }

    /** Calculates and sets the beam waist and the Rayleigh range (zR) from
     * the beam aperture (\f$ D_{\rm aperture} = \pi w_{\rm f} \f$) and the
     * focal length of a lens at the aperture. 
     *
     * \f$ w_0 = \frac{ f \lambda } { \pi w_{\rm f} } \f$.
     *
     * @param D_aperture
     *     The diameter of the limiting gaussian aperture
     *     (\f$ D_{\rm aperture} = \pi w_{\rm f} \f$) where \f$ w_{\rm f}\f$
     *     is the spot size at the aperture.
     * @param f
     *     The focal length of a lens placed at the aperture.
     */
    void setFocalProperties(const double & D_aperture,
                            const double & f) {
        w0 =     lambda *    (f / D_aperture);
        zR = 2 * lambda * SQR(f / D_aperture);
    }

    void setCollimation(const double & _w0 = -1.0) {
        if (_w0 > 0.0) w0 = _w0;
        zR = zR_collimated();
    }

    /** calculates and returns the intensity of the gaussian beam. */
    inline double operator()(const Vector<double,3> & r) const {
        if (!insideBeam(r)) return 0;

        register double w_z2_inverse = 1.0 / SQR(w(r));
        register double TwoRho2_wz2 = 2 * rho2(r) * w_z2_inverse;
        return P_0 * (2*fact(p)/ (physical::constant::pi * fact(abs(l) + p) ))
                   * w_z2_inverse
                   * fast_pow(TwoRho2_wz2, abs(l))
                   * SQR( LaguerreL(abs(l), p, TwoRho2_wz2) )
                   * exp(-TwoRho2_wz2);
    }

    /** Calculates the magnitude of the field.
     *   E(rho) = E_0 * exp( -rho^2 / w^2) * sqrt( w0 / w )
     */
    inline double E(const Vector<double,3> & r) const {
        return sqrt(operator()(r));
    }

    /** "Beam waist". */
    inline double w(const Vector<double,3> & r) const {
        return w0 * sqrt( 1.0 + SQR( z(r) / zR ) );
    }

    /** Radius of curvature. */
    inline double R(const Vector<double,3> & r) const {
        return z(r) * ( 1.0 + SQR( zR / z(r) ) );
    }

    /** Inside radius of cutoff. */
    inline bool insideBeam(const Vector<double,3> & r) const {
        return rho2(r) < cutoff_radius2;
    }

//  private:
    inline double rho2(const Vector<double,3> & r) const {
        return (SQR(r - r0) - SQR((r - r0)*k));
    }

    inline double z(const Vector<double,3> & r) const {
        return ((r - r0)*k);
    }

    inline double zR_collimated() const {
        return  physical::constant::pi * SQR(w0) / lambda;
    }
};

} /* namespace LaserCooling */

#endif // LAGUERREGAUSSIANLASER_H
