// -*- c++ -*-
// $Id$
/** \file
 * Convert a laser intensity function into an optical force.
 *
 * Copyright 2004-2007 Spencer Olson
 *
 */

#ifndef OPTICALFORCE_H
#define OPTICALFORCE_H

#include <olson-tools/physical.h>
#include <olson-tools/Forces.h>
#include <olson-tools/power.h>
#include <olson-tools/random/random.h>
#include <olson-tools/Distribution.h>

#include "AtomicResonance.h"
#include "Laser.h"

#define DO_SINGLE_SCATTERING
//#define DO_MULTIPLE_SCATTERING

#ifndef MULTILEVEL_ATOM
#  define TWO_LEVEL_ATOM
#endif

namespace LaserCooling {

/** Minimum number of spontaneous scatterings to occur before resorting to
 * average force+random walk.
 * It takes roughly 30 scatterings before the distributions in all three
 * dimensions converge to gaussian (according to the central limit theorem).
 */
const int MIN_SCAT = 30;

typedef struct {
    /** Return 0.5.
     */
    inline double distrib (const double & x) const {
        return 0.5;
    }
} FlatDistribution;

typedef struct {
    inline double distrib (const double & x) const {
        return SQR(cos(x));
    }
} P_THETA;

class NullSpontScatMonitor {
  public:
    /** For collecting information from single scattering events. */
    inline void addScatterSingle( Vector<double,6> & particle,
                                  const double & phi,
                                  const double & theta) const {}
    /** For collecting information from single scattering events. */
    inline void finishScatterSingle( Vector<double,6> & particle,
                                     const double & dt) const {}

    inline void addScatterDiffuse( const double & N_scat,
                                   Vector<double,6> & particle,
                                   const Vector<double,3> & dv,
                                   const double & dt) const {}
};

/** Calculate optical forces from a given laser field. 
 * @param Laser
 *     The laser field from which intensity and other parameters is drawn.
 * @param TYPEID
 *     TYPEID makes it possible to add two laser fields of a given Laser
 *     class; In other words, you should do: <br>
 *         typedef OpticalForce<Laser,0> beam1;
 *         typedef OpticalForce<Laser,1> beam2;
 *         AddForce< beam1, beam2 > LaserForce;
 * @param SpontScatMonitor
 *     An extension for monitoring the spontaneous scattering if the user so
 *     desires [Default does nothing].
 */
template <class Laser, unsigned int TYPEID = 0,
          class SpontScatMonitor = NullSpontScatMonitor>
class OpticalForce : public virtual BaseForce {
  public:
    typedef BaseForce super0;

    OpticalForce() :
        super0(), laser(), atom(), detuning(0.0),
        spontScatMonitor(),
#ifdef MULTILEVEL_ATOM
        g_p(1.0), Delta_LS(0.0),
#endif
        p_theta(P_THETA(),0,physical::constant::pi, 10000),
        p_phi(FlatDistribution(), 0, 2*physical::constant::pi, 2),
        p_x(GaussianDistrib(sqrt(20.0)),-2,2,10000),
        p_z(GaussianDistrib(sqrt(0.5)),-4,4,10000)
        {}

    /* **** STORAGE MEMBERS **** */
    Laser laser;
    AtomicResonance atom;

    /** Detuning away from the atomic resonance (Hz). */
    double detuning;

    /** A user-specified monitor for spontaneous scattering. */
    SpontScatMonitor spontScatMonitor;

#ifdef MULTILEVEL_ATOM
    /** Scattering rate per intensity (s^{-1} / (W/m^2)) */
    double g_p;

    /** Hyperfine splitting. 7.123 THz for Rb ( 5s -> 5p ).  */
    double Delta_LS;
#endif

  private:
    /* stuff for handling spontaneous emmission */
    SquareMatrix<double,3> R;
    Distribution p_theta, p_phi;
    Distribution p_x, p_z;

    /* we calculate these to speed things up later. */
    /** Recoil momentum [kg m/s]. */
    double p_recoil;

    /** Recoil velocity in m/s. */
    double v_recoil;


  public:
    /* **** FUNCTIONS *** */

    /** Copy operator.  Note that this does not set all of the private values.
     * Use initialize() after assigning with this function. */
    inline const OpticalForce & operator=(const OpticalForce & that) {
        super0::operator=(that);
        laser = that.laser;
        atom = that.atom;
        detuning = that.detuning;
        return *this;
    }

    /** Returns the Doppler shifted detuning (in Hz). */
    inline double Delta(const Vector<double,3> & v = V3(0,0,0)) const {
        return detuning - ((laser.k * v) / (2.0*physical::constant::pi));
    }

    inline double gamma_p(const Vector<double,3> & r,
                          const Vector<double,3> & v = V3(0,0,0)) const {
        /* FIXME:  let this function be templated in to correct for n-level
         * problem. */
#ifdef MULTILEVEL_ATOM
        /* From fredrik's calculations for lambda == 779.25. */
        return laser(r) * g_p /* (s^-1) / (W / m^2) */;
#else // TWO_LEVEL_ATOM
        /* The two level atom */
        register double s0 = laser(r) / atom.Isat();
        return      (s0 * physical::constant::pi * atom.Gamma())
                / ( 1.0 + s0 + SQR(2.0 * Delta(v) / atom.Gamma()) );
#endif
    }

    void initialize() {
        if (detuning == 0.0) {
            detuning = physical::constant::c * (1.0/laser.lambda - 1.0/atom.lambda());
        } else {
            laser.lambda = 1.0 / ( detuning / physical::constant::c + 1.0/atom.lambda() );
        }

	p_recoil = physical::constant::h / laser.lambda;
	v_recoil = p_recoil / super0::mass;

        /* I know that the method I am doing for this is not the fastest, but who
         * cares.  It is the most transparent and I only have to do this once
         * anyway!
         * */
        /* cos(phi) = x/sqrt(x**2 + y**2)
         * sin(phi) = y/sqrt(x**2 + y**2)
         * cos(theta (polar)) = z/r
         * sin(theta) = sqrt(1-cos(theta)**2)
         * */
        Vector<double,3> & k = laser.k;
        double costheta = k[Z] /* /r */;
        double sintheta = sqrt(1.0 - SQR(costheta));
        double cosphi = 1.0, sinphi = 0.0;
        if (sintheta) {
            cosphi = k[X] / hypot( k[X], k[Y]);
            sinphi = k[Y] / hypot( k[X], k[Y]);
        }

        /* rotation matrix is given by:
         * Rz[-phi].Ry[-theta]
         * where Rz[x] and Ry[x] are passive rotations.
         * */
        const double ry2_[3][3]  =
                             {{ costheta,    0,   sintheta},
                              {   0,         1,       0   },
                              {-sintheta,    0,   costheta}};

        const double rz2_[3][3]  =
                             {{ cosphi,   -sinphi,    0   },
                              { sinphi,    cosphi,    0   },
                              {   0,         0,       1   }};

        SquareMatrix<double,3> & ry2   = *((SquareMatrix<double,3>*)ry2_);
        SquareMatrix<double,3> & rz2   = *((SquareMatrix<double,3>*)rz2_);

        R   = rz2 * ry2;
    }

    /** Statistical force due to spontaneous scattering.
     * This function accounts for the random walk that occurs in momentum
     * space that affects velocity diffusion.
     */
    inline void applyStatisticalForce(Vector<double,6> & particle,
                                      const double & t, const double & dt) const {

        Vector<double,3> & x = V3C(particle.val);
        Vector<double,3> & v = V3C(particle.val + VX);
        Vector<double,3> dv;

        /* The factor of 0.5 is because this function is called before AND
         * after each iteration of the Runge-Kutta (rk5 adaptive).
         */
        register double N_scat = dt * gamma_p(x,v);

        /* for the following two blocks of this if-clause, dv will be computed
         * assuming the laser is pointing in the z-direction.  dv will then be
         * rotated back into the lab frame. */
#if defined(DO_SINGLE_SCATTERING)
        /* there seems to be a problem with doing this.  Hopefully the user
         * doesn't care about single scattering statistics. */
#  if defined(DO_MULTIPLE_SCATTERING)
        if (N_scat < MIN_SCAT)
#  endif
        {
            /* The scattering rate is so low that we will simulate each
             * scattering event individually.  We cannot assume the central
             * limit theorem.
             */
             int N_scat_for_sure = int(N_scat);
             N_scat -= N_scat_for_sure;

            if (MTRNGrand() <= N_scat ) {
                N_scat_for_sure++;
            }

            dv = 0.0;
            /* Now we do each of the N_scat_for_sure individual scatterings. */
            for (int i = 0; i < N_scat_for_sure; i++) {
                double phi   = p_phi();
                double theta = p_theta();
                dv[X] += cos(phi)*sin(theta);
                dv[Y] += sin(phi)*sin(theta);
                dv[Z] += cos(theta);

                spontScatMonitor.addScatterSingle(particle, phi, theta);
            }

            spontScatMonitor.finishScatterSingle(particle, dt);
        }
#  if defined(DO_MULTIPLE_SCATTERING)
        else
#  endif
#endif
#if defined(DO_MULTIPLE_SCATTERING)
        {
            /* The scattering rate is high enough that we just apply velocity
             * diffusion due to the random walk in momentum space.
             * Because of the central limit theorem, the scattered momentum
             * distributions are just gaussian:
             * If the laser is in the z-direction,
             * \f$ p(x) = p(y) \propto \exp\left( - x^2 \sqrt{20} / N_scat \right) \f$
             * and
             * \f$ p(z) \propto \exp\left( - z^2 \sqrt{1/2} / N_scat \right) \f$.
             *
             * Note that I did not actually do the convolution integrals, but
             * found this emperically.
             *
             * These results from sampling these distributions will have to be
             * changed back to the lab-frame coordinates. 
             */

            dv[X] = p_x();
            dv[Y] = p_x();
            dv[Z] = p_z();

            spontScatMonitor.addScatterDiffuse(N_scat, particle, dv, dt);

            dv *= sqrt(N_scat);
        }
#endif

        /* scale dv to the recoil velocity. */
        dv *= v_recoil;

        /* now we must rotate dv back into the lab frame. */
        v += R * dv;
    }

    /** Calculate the conservative dipole radiation force.  
     * The dipole force is given by:
     *
     * \f$ \vec{\bf F}_{\rm dip}
                       = \frac{\hbar \Delta}{2}
                         \frac{ \partial_{\vec{\bf x}} I(\vec{\bf x})/I_0 }
                              { 1 + (I(\vec{\bf x})/I_0)
                                  + \left(\frac{2\Delta}{\Gamma}\right)^2
                              }
       \f$
     */
    inline void F_dip(      Vector<double,3> & F,
                      const Vector<double,3> & r,
                      const Vector<double,3> & v = V3(0,0,0),
                      const double & t  = 0.0,
                      const double & dt = 0.0 ) const {
        /* FIXME: Note that I really do not need to calculate the detuning as
         * \f$ \delta - \vec{\bf k}\cdot\vec{\bf v} \f$ because when
         * \f$ |\delta| ~ \left|\vec{\bf k}\cdot\vec{\bf v}\right| \f$, this force is
         * negligable compared to the spontaneous emmision force.
         * I included \f$ \vec{\bf k}\cdot\vec{\bf v} \f$ for whatever wierd reason.
         */
        gradient(F,laser,r);
        register double _Delta = Delta(v);
#ifdef MULTILEVEL_ATOM
        F *= - ( physical::constant::h * SQR(atom.Gamma()) )
             / (
                 24.0 * atom.Isat()
                 * ( 2.0/_Delta + 1.0/(_Delta + Delta_LS) )
               );
#else // TWO_LEVEL_ATOM
        F *= - (0.5 * physical::constant::h * _Delta )
             / (
                 atom.Isat()
                 * (1.0 + (laser(r)/atom.Isat()) + SQR(2.0 * _Delta / atom.Gamma()))
               );
#endif
    }

    /** Calculate the average spontaneous radiation pressure force.  */
    inline void F_spont(      Vector<double,3> & F,
                        const Vector<double,3> & r,
                        const Vector<double,3> & v = V3(0,0,0),
                        const double & t = 0.0,
                        const double & dt = 0.0 ) const {
        register double _gamma_p = gamma_p(r,v);
#if defined(DO_SINGLE_SCATTERING)
        if ((dt * _gamma_p) < MIN_SCAT) {
            /* There are not enough scattering events to assume the central
             * limit theorem and apply an average force. */
            F = 0.0;
        } else
#endif
        {
            F = laser.k * (_gamma_p * p_recoil);
        }
    }

    /** Calculate the conservative optical potential.
     * The dipole potential is given by:
     *
     * \f$ U_{\rm dip} = \frac{\hbar \Delta}{2}
                         \log\right(
                            1 + \frac{(I(\vec{\bf x})/I_0)}
                                     {1 + \left(\frac{2\Delta}{\Gamma}\right)^2}
                         \left)
       \f$
     */
    inline double U_dip(const Vector<double,3> & r,
                        const Vector<double,3> & v = V3(0,0,0),
                        const double & t = 0.0) const {
        register double _Delta = Delta(v);

#ifdef MULTILEVEL_ATOM
        return laser(r)
             * ( physical::constant::h * SQR(atom.Gamma()) )
             / (
                 24.0 * atom.Isat()
                 * ( 2.0/_Delta + 1.0/(_Delta + Delta_LS) )
               );
#else // TWO_LEVEL_ATOM
        return
              (0.5 * physical::constant::h * _Delta)
            * log( 1.0
                    +     laser(r)
                      / ( atom.Isat()
                          * (1.0 + SQR(2.0 * _Delta / atom.Gamma()))
                        )
                 );
#endif
    }


    /* ************ Force template necessecities. ********** */

    /** Calculate the conservative dipole radiation force.
     * NOTE:  This function might be way too slow and overkill because it is
     * trying to generalize to all detunings.  There are situations where
     * F_dip can be negelcted and times when F_spont might be neglected.
     */
    inline void accel(      Vector<double,3> & a,
                      const Vector<double,3> & r,
                      const Vector<double,3> & v = V3(0,0,0),
                      const double & t  = 0.0,
                      const double & dt = 0.0) const {
        F_dip(a,r,v,t,dt);

#if defined(DO_MULTIPLE_SCATTERING)
        Vector<double,3> Fsp;
        F_spont(Fsp,r,v,t,dt);

        a += Fsp;
#endif
        a *= 1.0 / super0::mass;
    }

    /** Calculate the conservative optical potential ignoring the
     * gravitational component.
     * @see potential(const Vector<double,3> & r).
     */
    inline double potential(const Vector<double,3> & r,
                            const Vector<double,3> & v = V3(0,0,0),
                            const double & t = 0.0) const {
        return U_dip(r,v,t);
    }

};

} /* namespace LaserCooling */

#endif // OPTICALFORCE_H
