// -*- c++ -*-
// $Id$
/** \file
 * Monitor spontaneous scattering and propagate quantum tragectories as a
 * result..
 *
 * Copyright 2004-2007 Spencer Olson
 *
 */


#ifndef SPONTSCATMON_H
#define SPONTSCATMON_H

#include <olson-tools/GenericBin.h>
#include <olson-tools/GenericBinExtender.h>
#include "StateParticle.h"

namespace LaserCooling {


enum SCAT_BIN_ELT {
    SCAT_N_TEST = 0,
    SCAT_N_SCAT = 1,
    SCAT_DT_TOT = 2
};

typedef GenericBin<double,20, Vector<double,3> > ZBin;

typedef GenericBinExtender<double, ZBin, 20> YBin;

GenericBinExtender<double, YBin, 20> SScatBin;
double total_scattering;
int total_scatter_tests;
double total_scattering_time;

void initSScatBin() {
#if 0
    const Vector<double,3> mn(VInit, -2*mm, -100*um, -100*um);
    const Vector<double,3> mx(VInit,  2*mm,  100*um,  100*um);
#else
    const Vector<double,3> mn(VInit, -2*mm, -2*mm, -2*mm);
    const Vector<double,3> mx(VInit,  2*mm,  2*mm,  2*mm);
#endif

    SScatBin.init(mn[X],mx[X]);
    SScatBin.initChildBins(mn[Y],mx[Y]);
    for(unsigned int i = 0;
        i < SScatBin.nBins();
        SScatBin.bins[i++].initChildBins(mn[Z],mx[Z]));

    total_scattering = 0.0;
    total_scatter_tests = 0;
    total_scattering_time = 0.0;
}

class SpontScatMonitor {
  public:
    double g23, g32;
    SpontScatMonitor() : g32(0.5), g23(0.5) {}
    /** For collecting information from single scattering events. */
    inline void addScatterSingle( Vector<double,6> & particle,
                                  const double & phi, const double & theta) const {
        StateParticle & p = (StateParticle&)particle;
        calculateQuantumTrajectory(1, p);

        /* Second, we take the time to record the scattering at this location.
         * */
        Vector<double,3> & vec = SScatBin.getBin(p.pos[X]).getBin(p.pos[Y]).getBin(p.pos[Z]);
        vec[SCAT_N_SCAT] += 1;
        total_scattering += 1;
    }

    /** For collecting information from single scattering events. */
    inline void finishScatterSingle( Vector<double,6> & particle,
                                     const double & dt ) const {
        StateParticle & p = (StateParticle&)particle;

        /* Second, we take the time to record the scattering at this location.
         * */
        Vector<double,3> & vec = SScatBin.getBin(p.pos[X]).getBin(p.pos[Y]).getBin(p.pos[Z]);
        vec[SCAT_N_TEST] += 1;
        vec[SCAT_DT_TOT] += dt;
        total_scatter_tests++;
        total_scattering_time += dt;
    }

    inline void addScatterDiffuse( const double & N_scat,
                                   Vector<double,6> & particle,
                                   const Vector<double,3> & dv,
                                   const double & dt ) const {
        StateParticle & p = (StateParticle&)particle;

        int N_change_for_sure = int(N_scat);
        double N_possible_change = N_scat - N_change_for_sure;

        if (MTRNGrand() <= N_possible_change ) {
            N_change_for_sure++;
        }

        calculateQuantumTrajectory(N_change_for_sure, p);

        /* Second, we take the time to record the scattering at this location.
         * */
        Vector<double,3> & vec = SScatBin.getBin(p.pos[X]).getBin(p.pos[Y]).getBin(p.pos[Z]);
        vec[SCAT_N_SCAT] += N_scat;
        vec[SCAT_N_TEST] += 1.0;
        vec[SCAT_DT_TOT] += dt;
        total_scattering += N_scat;
        total_scatter_tests++;
        total_scattering_time += dt;
    }

  private:
    void calculateQuantumTrajectory(int N_change_for_sure, StateParticle & p) const {
        /* first, we take this time to calculate any state changes for the
         * particle.
         * If N_scat < 1, then we just do a single quantum trajectory.
         *     We also have to pick, based on branching ratios, which state to
         *     enter.
         * If N_scat >= 1, then we do (int(N_scat)) scatterings for sure where
         *     we pick the successive states based on branching ratios.
         *     The remaining fraction is calculated via the (N_scat<1) method.
         */
        /* Now we do each of the N_scat_for_sure individual scatterings. */
        for (; N_change_for_sure > 0; N_change_for_sure--) {
            if (p.state == StateParticle::F2) {
                if (MTRNGrand() < g23) {
                    p.state = StateParticle::F3;
                }
            } else { /* state == F3 */
                if (MTRNGrand() < g32) {
                    p.state = StateParticle::F2;
                }
            }
        }/*for*/
    }
};

} /* namespace LaserCooling */

#endif // SPONTSCATMON_H
