// -*- c++ -*-
// $Id$
/** \file
 * Particle class that maintains state information.
 *
 * Copyright 2000-2007 Spencer Olson
 *
 */

#ifndef STATE_PARTICLE_H
#define STATE_PARTICLE_H

#include <dsmc/Particle.h>


/** To work well with the DSMC stuff, I think that I have to have similar
 * constructors to the base Particle. */
class StateParticle : public ParticleBase {
  public:
    /** Pointer to next Particle in a linked list of Particle\s.
     */
    StateParticle * next;

    /** we redef the new and delete operators so that we can cache these for
     * multiple tree building.
     * Note:  it is important that this is declared on the latter end of the
     * object so that passing a Particle to code written in non-c++ (eg.
     * fortran) will work.
     */
    REDEF_NEW_DEL_H(StateParticle);
  public:

    enum STATE {
        F2,
        F3
    };

    enum STATE state;

    inline StateParticle(const int & typ = 0, const double & dt_init = 0,
                         const enum STATE & _state = F2 ):
        ParticleBase(typ,dt_init), next(NULL), state(_state) { }

    /** Copy contructor--does not copy the next field. */
    inline StateParticle(const StateParticle & that) : next(NULL) { *this = that; }

    /** Copy operator--does NOT change the next field. */
    inline const StateParticle & operator=(const StateParticle & that) {
        register StateParticle * ptr = next;
        ParticleBase::operator=(that);
        next = ptr;
        state = that.state;
        return *this;
    }
}; // class StateParticle

/** Print the six position and velocity values out seperated by spaces.
 */
inline std::ostream & operator<<(std::ostream & output,const StateParticle & q) {
    return output << ((const Particle&)q) << ' ' << ((int)q.state);
}

/** Read in the six position and velocity values and the state information
 * into a newly created particle.
 * This function satisfies the requirements for the readData routines in the
 * olson-tools package.
 */
std::istream & operator>>(std::istream & input, StateParticle *& p);

#  if defined(USE_MPI)

#include <dsmc/ParticleNode.h>

/** StateParticle size specialization for ParticleNode (MPI) use. */
template<>
inline int particle_packed_size<StateParticle>(MPI::Comm & comm) {
    return
        particle_packed_size<ParticleBase>(comm)
        + MPI::INT.Pack_size(1,comm);
}

/** StateParticle unpack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_unpack<StateParticle>(const uint8_t * buffer,
                                           const int & bufsz,
                                           StateParticle * p,
                                           int & bpos,
                                           MPI::Comm & comm ) {
    /* unpack Particle and state */
    particle_unpack<ParticleBase>(buffer, bufsz, p, bpos, comm);
    MPI::INT.Unpack(buffer, bufsz, &p->state, 1, bpos, comm);
}

/** StateParticle pack specialization for ParticleNode (MPI) use. */
template<>
inline void particle_pack<StateParticle>(uint8_t * buffer,
                                         const int & bufsz,
                                         const StateParticle * p,
                                         int & bpos,
                                         MPI::Comm & comm) {
    /* Particle */
    particle_pack<ParticleBase>(buffer, bufsz, p, bpos, comm);
    /* state */
    MPI::INT.Pack(&p->state, 1, buffer, bufsz, bpos, comm);
}

#  endif

#endif // STATE_PARTICLE_H
