// -*- c++ -*-
// $Id$
/** \file
 * Laser field lookup specialization of FieldLookup class.
 *
 * Copyright 2005-2007 Spencer Olson
 *
 * @see createFieldFile.h for routines to help creating the field-lookup table
 * file.
 */



#ifndef __LASERFIELD_LOOKUP_H
#define __LASERFIELD_LOOKUP_H

#include <iostream>
#include <string>
#include <sstream>
#include <stdexcept>
#include <olson-tools/ompexcept.h>
#include <olson-tools/field-lookup.h>

namespace LaserCooling {


/* ************** BEGIN FORCE LOOKUP SPECIALIZATION *********** */

/** Intensity record for lookup table.
 * @see ForceRecord.
 */
template <unsigned int L = 3U>
class IntensityRecord {
  public:
    IntensityRecord() : I(0.0) {}
    double I;

    /** For using the FieldLookup::scalar_lookup routine. */
    inline double & scalar(const unsigned int & i) { return I; }

    /** For using the FieldLookup::scalar_lookup routine. */
    inline const double & scalar(const unsigned int & i) const { return I; }
};

template <unsigned int L>
inline std::istream & operator>>(std::istream & input, IntensityRecord<L> & Ir) {
    Vector<double,L> position;
    input >> position >> Ir.I;
    return input;
}

template <unsigned int L>
inline std::ostream & operator<<(std::ostream & output, const IntensityRecord<L> & Ir) {
    output << Ir.I;
    return output;
}

/** The base Lookup class of the associated FieldLookup container class.
 * If you have an axially symmetric force, you can more optimally use
 * AxiSymFieldLookup< IntensityRecord<L> >.  This will both speed up the lookup
 * process as well as minimize the memory footprint.
 *
 * @see FieldLookup for cartesian lookup table.
 * @see AxiSymFieldLookup for axially symmetric lookup table.
 */
template <unsigned int L = 3U, class T = FieldLookup< IntensityRecord<L> > >
class LaserFieldLookup : public virtual BaseField, public T {
  public:
    typedef BaseField super0;
    typedef T super1;

    /** Default constructor. */
    LaserFieldLookup(const double & _PowerFactor = physical::units::mW,
                     const double & _polarization  = 1,
                     const Vector<double,3> _k  = V3(1,0,0),
                     const double & _lambda     = 780 * physical::units::nm ) :
        super0(),
        super1(),
        P(_PowerFactor),
        PowerFactor(_PowerFactor),
        polarization(_polarization),
        k(_k/_k.abs()),
        lambda(_lambda) {
    }

    double P;                   /* Power (W). */
    double PowerFactor;         /* Power / (Integral of laser intensity transverse to k). */
    double polarization;        /* Polarization (+1 or -1; see above explanation. */
    Vector<double,3> k;         /* wave number (normalized) */
    double lambda;              /* wavelength of laser (m) */

    /** Compute/lookup the intensity of an optical source.
     * @param r
     *     The position to evaluate the field.
     */
    inline double operator()(const Vector<double,3> & r) const {
        return PowerFactor * super1::scalar_lookup(r, 0);
    }

    void setPowerFactor(const double & Power, const double & Integral ) {
        PowerFactor = Power / Integral;
        P = Power;
    }

}; /* LaserFieldLookup class */

template <class T, unsigned int L = 3U>
class IntensityTableWrapper : public T {
  public:
    typedef T super;
    IntensityRecord<L> getRecord(const Vector<double,3> & r) const {
        IntensityRecord<L> retval;
        retval.V = super::operator()(r);
        return retval;
    }
};

/* **************   END FORCE LOOKUP SPECIALIZATION *********** */

} /* namespace LaserCooling */

#endif // __LASERFIELD_LOOKUP_H
